@@ -1,26 +1,30 @@
 
 /*
- * Copyright (C) 2011 Mail.RU
+ * Redistribution and use in source and binary forms, with or
+ * without modification, are permitted provided that the following
+ * conditions are met:
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
+ * 1. Redistributions of source code must retain the above
+ *    copyright notice, this list of conditions and the
+ *    following disclaimer.
  *
- * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * 2. Redistributions in binary form must reproduce the above
+ *    copyright notice, this list of conditions and the following
+ *    disclaimer in the documentation and/or other materials
+ *    provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
+ * <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
+ * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+ * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
+ * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  */
 
@@ -45,7 +49,7 @@ struct tnt_sql {
 };
 
 static bool
-tnt_sql_error(struct tnt_sql *sql, struct tnt_tk *last, char *fmt, ...)
+tnt_sql_error(struct tnt_sql *sql, struct tnt_tk *last, const char *fmt, ...)
 {
 	char msgu[256];
 	va_list args;
@@ -73,7 +77,7 @@ tnt_sql_tk(struct tnt_sql *sql, int tk, struct tnt_tk **tkp)
 	if (tk_ != tk) {
 		if (tk < 0xff && ispunct(tk))
 			return tnt_sql_error(sql, tkp_, "expected '%c'", tk);
-		return tnt_sql_error(sql, tkp_, "expected '%s'", tnt_lex_nameof(tk));
+		return tnt_sql_error(sql, tkp_, "expected '%s'", tnt_lex_nameof(sql->l, tk));
 	}
 	if (tkp)
 		*tkp = tkp_;
@@ -178,7 +182,7 @@ tnt_sql_stmt_update(struct tnt_sql *sql, struct tnt_tuple *tu, struct tnt_stream
 	tnt_expect(tnt_sqltkv(sql, TNT_TK_TABLE, &tn));
 	tnt_expect(tnt_sqltk(sql, TNT_TK_SET));
 	while (1) {
-		struct tnt_tk *k;
+		struct tnt_tk *k = NULL;
 		tnt_expect(tnt_sqltkv(sql, TNT_TK_KEY, &k));
 		tnt_expect(tnt_sqltk(sql, '='));
 		struct tnt_tk *v;
@@ -224,7 +228,8 @@ tnt_sql_stmt_update(struct tnt_sql *sql, struct tnt_tuple *tu, struct tnt_stream
 			break;
 		/* k = splice(k, a, b) */
 		case TNT_TK_SPLICE: {
-			struct tnt_tk *field, *off, *len, *list;
+			struct tnt_tk *field = NULL, *off = NULL,
+				      *len = NULL, *list = NULL;
 			tnt_expect(tnt_sqltk(sql, '('));
 			tnt_expect(tnt_sqltkv(sql, TNT_TK_KEY, &field));
 			if (TNT_TK_I32(k) != TNT_TK_I32(field)) {
@@ -370,7 +375,7 @@ tnt_sql_stmt(struct tnt_sql *sql)
 		char proc[512];
 		int len = 0;
 		while (1) {
-			struct tnt_tk *name;
+			struct tnt_tk *name = NULL;
 			tnt_lex_idonly(sql->l, true);
 			tnt_expect(tnt_sqltkv(sql, TNT_TK_ID, &name));
 			tnt_lex_idonly(sql->l, false);
@@ -436,6 +441,27 @@ static bool tnt_sql(struct tnt_sql *sql) {
 	return tnt_sql_stmt(sql);
 }
 
+struct tnt_lex_keyword tnt_sql_keywords[] =
+{
+	{  "PING",    4, TNT_TK_PING },
+	{  "UPDATE",  6, TNT_TK_UPDATE },
+	{  "SET",     3, TNT_TK_SET },
+	{  "WHERE",   5, TNT_TK_WHERE },
+	{  "SPLICE",  6, TNT_TK_SPLICE },
+	{  "DELETE",  6, TNT_TK_DELETE },
+	{  "FROM",    4, TNT_TK_FROM },
+	{  "INSERT",  6, TNT_TK_INSERT },
+	{  "REPLACE", 7, TNT_TK_REPLACE },
+	{  "INTO",    4, TNT_TK_INTO },
+	{  "VALUES",  6, TNT_TK_VALUES },
+	{  "SELECT",  6, TNT_TK_SELECT },
+	{  "OR",      2, TNT_TK_OR },
+	{  "AND",     3, TNT_TK_AND },
+	{  "LIMIT",   5, TNT_TK_LIMIT },
+	{  "CALL",    4, TNT_TK_CALL },
+	{  NULL,      0, TNT_TK_NONE }
+};
+
 /*
  * tnt_query()
  *
@@ -450,10 +476,10 @@ static bool tnt_sql(struct tnt_sql *sql) {
  * and string description returned (must be freed after use).
 */
 int
-tnt_query(struct tnt_stream *s, char *q, size_t qsize, char **e)
+tnt_query(struct tnt_stream *s, const char *q, size_t qsize, char **e)
 {
 	struct tnt_lex l;
-	if (!tnt_lex_init(&l, (unsigned char*)q, qsize))
+	if (!tnt_lex_init(&l, tnt_sql_keywords, (unsigned char*)q, qsize))
 		return -1;
 	struct tnt_sql sql = { s, &l, NULL };
 	bool ret = tnt_sql(&sql);
@@ -481,7 +507,7 @@ int
 tnt_query_is(char *q, size_t qsize)
 {
 	struct tnt_lex l;
-	if (!tnt_lex_init(&l, (unsigned char*)q, qsize))
+	if (!tnt_lex_init(&l, tnt_sql_keywords, (unsigned char*)q, qsize))
 		return 0;
 	int rc = 0;
 	struct tnt_tk *tk;
