@@ -1,31 +1,33 @@
 
 #line 1 "src/admin.rl"
 /*
- * Copyright (C) 2010 Mail.RU
- * Copyright (C) 2010 Yuriy Vostrikov
+ * Redistribution and use in source and binary forms, with or
+ * without modification, are permitted provided that the following
+ * conditions are met:
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
+ * 1. Redistributions of source code must retain the above
+ *    copyright notice, this list of conditions and the
+ *    following disclaimer.
  *
- * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * 2. Redistributions in binary form must reproduce the above
+ *    copyright notice, this list of conditions and the following
+ *    disclaimer in the documentation and/or other materials
+ *    provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
+ * <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
+ * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+ * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
+ * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  */
-
 #include <stdio.h>
 #include <string.h>
 #include <stdbool.h>
@@ -38,16 +40,17 @@
 #include <say.h>
 #include <stat.h>
 #include <tarantool.h>
-#include <tarantool_lua.h>
+#include "lua/init.h"
 #include <recovery.h>
-#include TARANTOOL_CONFIG
 #include <tbuf.h>
 #include <util.h>
 #include <errinj.h>
+#include "coio_buf.h"
 
 #include "lua.h"
 #include "lauxlib.h"
 #include "lualib.h"
+#include "box/box.h"
 
 static const char *help =
 	"available commands:" CRLF
@@ -69,7 +72,7 @@ static const char *help =
 static const char *unknown_command = "unknown command. try typing help." CRLF;
 
 
-#line 73 "src/admin.m"
+#line 76 "src/admin.m"
 static const int admin_start = 1;
 static const int admin_first_final = 135;
 static const int admin_error = 0;
@@ -77,9 +80,52 @@ static const int admin_error = 0;
 static const int admin_en_main = 1;
 
 
-#line 72 "src/admin.rl"
+#line 75 "src/admin.rl"
 
 
+struct salloc_stat_admin_cb_ctx {
+	i64 total_used;
+	struct tbuf *out;
+};
+
+static int
+salloc_stat_admin_cb(const struct slab_class_stats *cstat, void *cb_ctx)
+{
+	struct salloc_stat_admin_cb_ctx *ctx = cb_ctx;
+
+	tbuf_printf(ctx->out,
+		    "     - { item_size: %- 5i, slabs: %- 3i, items: %- 11" PRIi64
+		    ", bytes_used: %- 12" PRIi64
+		    ", bytes_free: %- 12" PRIi64 " }" CRLF,
+		    (int)cstat->item_size,
+		    (int)cstat->slabs,
+		    cstat->items,
+		    cstat->bytes_used,
+		    cstat->bytes_free);
+
+	ctx->total_used += cstat->bytes_used;
+	return 0;
+}
+
+static void
+show_slab(struct tbuf *out)
+{
+	struct salloc_stat_admin_cb_ctx cb_ctx;
+	struct slab_arena_stats astat;
+
+	cb_ctx.total_used = 0;
+	cb_ctx.out = out;
+
+	tbuf_printf(out, "slab statistics:\n  classes:" CRLF);
+
+	salloc_stat(salloc_stat_admin_cb, &astat, &cb_ctx);
+
+	tbuf_printf(out, "  items_used: %.2f%%" CRLF,
+		(double)cb_ctx.total_used / astat.size * 100);
+	tbuf_printf(out, "  arena_used: %.2f%%" CRLF,
+		(double)astat.used / astat.size * 100);
+}
+
 
 static void
 end(struct tbuf *out)
@@ -117,6 +163,7 @@ tarantool_info(struct tbuf *out)
 	tbuf_printf(out, "  uptime: %i" CRLF, (int)tarantool_uptime());
 	tbuf_printf(out, "  pid: %i" CRLF, getpid());
 	tbuf_printf(out, "  logger_pid: %i" CRLF, logger_pid);
+	tbuf_printf(out, "  snapshot_pid: %i" CRLF, snapshot_pid);
 	tbuf_printf(out, "  lsn: %" PRIi64 CRLF,
 		    recovery_state->confirmed_lsn);
 	tbuf_printf(out, "  recovery_lag: %.3f" CRLF,
@@ -125,7 +172,7 @@ tarantool_info(struct tbuf *out)
 	tbuf_printf(out, "  recovery_last_update: %.3f" CRLF,
 		    recovery_state->remote ?
 		    recovery_state->remote->recovery_last_update_tstamp :0);
-	mod_info(out);
+	box_info(out);
 	const char *path = cfg_filename_fullpath;
 	if (path == NULL)
 		path = cfg_filename;
@@ -133,8 +180,27 @@ tarantool_info(struct tbuf *out)
 }
 
 static int
-admin_dispatch(lua_State *L)
+show_stat_item(const char *name, int rps, i64 total, void *ctx)
 {
+	struct tbuf *buf = ctx;
+	int name_len = strlen(name);
+	tbuf_printf(buf,
+		    "  %s:%*s{ rps: %- 6i, total: %- 12" PRIi64 " }" CRLF,
+		    name, 1 + stat_max_name_len - name_len, " ", rps, total);
+	return 0;
+}
+
+void
+show_stat(struct tbuf *buf)
+{
+	tbuf_printf(buf, "statistics:" CRLF);
+	stat_foreach(show_stat_item, buf);
+}
+
+static int
+admin_dispatch(struct ev_io *coio, struct iobuf *iobuf, lua_State *L)
+{
+	struct ibuf *in = &iobuf->in;
 	struct tbuf *out = tbuf_alloc(fiber->gc_pool);
 	struct tbuf *err = tbuf_alloc(fiber->gc_pool);
 	int cs;
@@ -142,21 +208,21 @@ admin_dispatch(lua_State *L)
 	char *strstart, *strend;
 	bool state;
 
-	while ((pe = memchr(fiber->rbuf->data, '\n', fiber->rbuf->size)) == NULL) {
-		if (fiber_bread(fiber->rbuf, 1) <= 0)
-			return 0;
+	while ((pe = memchr(in->pos, '\n', in->end - in->pos)) == NULL) {
+		if (coio_bread(coio, in, 1) <= 0)
+			return -1;
 	}
 
 	pe++;
-	p = fiber->rbuf->data;
+	p = in->pos;
 
 	
-#line 155 "src/admin.m"
+#line 221 "src/admin.m"
 	{
 	cs = admin_start;
 	}
 
-#line 160 "src/admin.m"
+#line 226 "src/admin.m"
 	{
 	if ( p == pe )
 		goto _test_eof;
@@ -219,15 +285,15 @@ case 6:
 	}
 	goto st0;
 tr13:
-#line 254 "src/admin.rl"
+#line 308 "src/admin.rl"
 	{slab_validate(); ok(out);}
 	goto st135;
 tr20:
-#line 242 "src/admin.rl"
-	{return 0;}
+#line 296 "src/admin.rl"
+	{return -1;}
 	goto st135;
 tr25:
-#line 169 "src/admin.rl"
+#line 223 "src/admin.rl"
 	{
 			start(out);
 			tbuf_append(out, help, strlen(help));
@@ -235,9 +301,9 @@ tr25:
 		}
 	goto st135;
 tr36:
-#line 228 "src/admin.rl"
+#line 282 "src/admin.rl"
 	{strend = p;}
-#line 175 "src/admin.rl"
+#line 229 "src/admin.rl"
 	{
 			strstart[strend-strstart]='\0';
 			start(out);
@@ -246,7 +312,7 @@ tr36:
 		}
 	goto st135;
 tr43:
-#line 182 "src/admin.rl"
+#line 236 "src/admin.rl"
 	{
 			if (reload_cfg(err))
 				fail(out, err);
@@ -255,11 +321,11 @@ tr43:
 		}
 	goto st135;
 tr67:
-#line 252 "src/admin.rl"
+#line 306 "src/admin.rl"
 	{coredump(60); ok(out);}
 	goto st135;
 tr76:
-#line 189 "src/admin.rl"
+#line 243 "src/admin.rl"
 	{
 			int ret = snapshot(NULL, 0);
 
@@ -274,9 +340,9 @@ tr76:
 		}
 	goto st135;
 tr98:
-#line 238 "src/admin.rl"
+#line 292 "src/admin.rl"
 	{ state = false; }
-#line 202 "src/admin.rl"
+#line 256 "src/admin.rl"
 	{
 			strstart[strend-strstart] = '\0';
 			if (errinj_set_byname(strstart, state)) {
@@ -288,9 +354,9 @@ tr98:
 		}
 	goto st135;
 tr101:
-#line 237 "src/admin.rl"
+#line 291 "src/admin.rl"
 	{ state = true; }
-#line 202 "src/admin.rl"
+#line 256 "src/admin.rl"
 	{
 			strstart[strend-strstart] = '\0';
 			if (errinj_set_byname(strstart, state)) {
@@ -302,35 +368,23 @@ tr101:
 		}
 	goto st135;
 tr117:
-#line 145 "src/admin.rl"
+#line 211 "src/admin.rl"
 	{
-			tarantool_cfg_iterator_t *i;
-			char *key, *value;
-
 			start(out);
-			tbuf_printf(out, "configuration:" CRLF);
-			i = tarantool_cfg_iterator_init();
-			while ((key = tarantool_cfg_iterator_next(i, &cfg, &value)) != NULL) {
-				if (value) {
-					tbuf_printf(out, "  %s: \"%s\"" CRLF, key, value);
-					free(value);
-				} else {
-					tbuf_printf(out, "  %s: (null)" CRLF, key);
-				}
-			}
+			show_cfg(out);
 			end(out);
 		}
 	goto st135;
 tr131:
-#line 245 "src/admin.rl"
+#line 299 "src/admin.rl"
 	{start(out); fiber_info(out); end(out);}
 	goto st135;
 tr137:
-#line 244 "src/admin.rl"
+#line 298 "src/admin.rl"
 	{start(out); tarantool_info(out); end(out);}
 	goto st135;
 tr146:
-#line 163 "src/admin.rl"
+#line 217 "src/admin.rl"
 	{
 			start(out);
 			errinj_info(out);
@@ -338,33 +392,33 @@ tr146:
 		}
 	goto st135;
 tr152:
-#line 248 "src/admin.rl"
+#line 302 "src/admin.rl"
 	{start(out); palloc_stat(out); end(out);}
 	goto st135;
 tr160:
-#line 247 "src/admin.rl"
-	{start(out); slab_stat(out); end(out);}
+#line 301 "src/admin.rl"
+	{start(out); show_slab(out); end(out);}
 	goto st135;
 tr164:
-#line 249 "src/admin.rl"
-	{start(out); stat_print(out);end(out);}
+#line 303 "src/admin.rl"
+	{start(out); show_stat(out);end(out);}
 	goto st135;
 st135:
 	if ( ++p == pe )
 		goto _test_eof135;
 case 135:
-#line 357 "src/admin.m"
+#line 411 "src/admin.m"
 	goto st0;
 tr14:
-#line 254 "src/admin.rl"
+#line 308 "src/admin.rl"
 	{slab_validate(); ok(out);}
 	goto st7;
 tr21:
-#line 242 "src/admin.rl"
-	{return 0;}
+#line 296 "src/admin.rl"
+	{return -1;}
 	goto st7;
 tr26:
-#line 169 "src/admin.rl"
+#line 223 "src/admin.rl"
 	{
 			start(out);
 			tbuf_append(out, help, strlen(help));
@@ -372,9 +426,9 @@ tr26:
 		}
 	goto st7;
 tr37:
-#line 228 "src/admin.rl"
+#line 282 "src/admin.rl"
 	{strend = p;}
-#line 175 "src/admin.rl"
+#line 229 "src/admin.rl"
 	{
 			strstart[strend-strstart]='\0';
 			start(out);
@@ -383,7 +437,7 @@ tr37:
 		}
 	goto st7;
 tr44:
-#line 182 "src/admin.rl"
+#line 236 "src/admin.rl"
 	{
 			if (reload_cfg(err))
 				fail(out, err);
@@ -392,11 +446,11 @@ tr44:
 		}
 	goto st7;
 tr68:
-#line 252 "src/admin.rl"
+#line 306 "src/admin.rl"
 	{coredump(60); ok(out);}
 	goto st7;
 tr77:
-#line 189 "src/admin.rl"
+#line 243 "src/admin.rl"
 	{
 			int ret = snapshot(NULL, 0);
 
@@ -411,9 +465,9 @@ tr77:
 		}
 	goto st7;
 tr99:
-#line 238 "src/admin.rl"
+#line 292 "src/admin.rl"
 	{ state = false; }
-#line 202 "src/admin.rl"
+#line 256 "src/admin.rl"
 	{
 			strstart[strend-strstart] = '\0';
 			if (errinj_set_byname(strstart, state)) {
@@ -425,9 +479,9 @@ tr99:
 		}
 	goto st7;
 tr102:
-#line 237 "src/admin.rl"
+#line 291 "src/admin.rl"
 	{ state = true; }
-#line 202 "src/admin.rl"
+#line 256 "src/admin.rl"
 	{
 			strstart[strend-strstart] = '\0';
 			if (errinj_set_byname(strstart, state)) {
@@ -439,35 +493,23 @@ tr102:
 		}
 	goto st7;
 tr118:
-#line 145 "src/admin.rl"
+#line 211 "src/admin.rl"
 	{
-			tarantool_cfg_iterator_t *i;
-			char *key, *value;
-
 			start(out);
-			tbuf_printf(out, "configuration:" CRLF);
-			i = tarantool_cfg_iterator_init();
-			while ((key = tarantool_cfg_iterator_next(i, &cfg, &value)) != NULL) {
-				if (value) {
-					tbuf_printf(out, "  %s: \"%s\"" CRLF, key, value);
-					free(value);
-				} else {
-					tbuf_printf(out, "  %s: (null)" CRLF, key);
-				}
-			}
+			show_cfg(out);
 			end(out);
 		}
 	goto st7;
 tr132:
-#line 245 "src/admin.rl"
+#line 299 "src/admin.rl"
 	{start(out); fiber_info(out); end(out);}
 	goto st7;
 tr138:
-#line 244 "src/admin.rl"
+#line 298 "src/admin.rl"
 	{start(out); tarantool_info(out); end(out);}
 	goto st7;
 tr147:
-#line 163 "src/admin.rl"
+#line 217 "src/admin.rl"
 	{
 			start(out);
 			errinj_info(out);
@@ -475,22 +517,22 @@ tr147:
 		}
 	goto st7;
 tr153:
-#line 248 "src/admin.rl"
+#line 302 "src/admin.rl"
 	{start(out); palloc_stat(out); end(out);}
 	goto st7;
 tr161:
-#line 247 "src/admin.rl"
-	{start(out); slab_stat(out); end(out);}
+#line 301 "src/admin.rl"
+	{start(out); show_slab(out); end(out);}
 	goto st7;
 tr165:
-#line 249 "src/admin.rl"
-	{start(out); stat_print(out);end(out);}
+#line 303 "src/admin.rl"
+	{start(out); show_stat(out);end(out);}
 	goto st7;
 st7:
 	if ( ++p == pe )
 		goto _test_eof7;
 case 7:
-#line 494 "src/admin.m"
+#line 536 "src/admin.m"
 	if ( (*p) == 10 )
 		goto st135;
 	goto st0;
@@ -643,28 +685,28 @@ case 23:
 	}
 	goto tr33;
 tr33:
-#line 228 "src/admin.rl"
+#line 282 "src/admin.rl"
 	{strstart = p;}
 	goto st24;
 st24:
 	if ( ++p == pe )
 		goto _test_eof24;
 case 24:
-#line 654 "src/admin.m"
+#line 696 "src/admin.m"
 	switch( (*p) ) {
 		case 10: goto tr36;
 		case 13: goto tr37;
 	}
 	goto st24;
 tr34:
-#line 228 "src/admin.rl"
+#line 282 "src/admin.rl"
 	{strstart = p;}
 	goto st25;
 st25:
 	if ( ++p == pe )
 		goto _test_eof25;
 case 25:
-#line 668 "src/admin.m"
+#line 710 "src/admin.m"
 	switch( (*p) ) {
 		case 10: goto tr36;
 		case 13: goto tr37;
@@ -1114,28 +1156,28 @@ case 73:
 		goto tr91;
 	goto st0;
 tr91:
-#line 236 "src/admin.rl"
+#line 290 "src/admin.rl"
 	{ strstart = p; }
 	goto st74;
 st74:
 	if ( ++p == pe )
 		goto _test_eof74;
 case 74:
-#line 1125 "src/admin.m"
+#line 1167 "src/admin.m"
 	if ( (*p) == 32 )
 		goto tr92;
 	if ( 33 <= (*p) && (*p) <= 126 )
 		goto st74;
 	goto st0;
 tr92:
-#line 236 "src/admin.rl"
+#line 290 "src/admin.rl"
 	{ strend = p; }
 	goto st75;
 st75:
 	if ( ++p == pe )
 		goto _test_eof75;
 case 75:
-#line 1139 "src/admin.m"
+#line 1181 "src/admin.m"
 	switch( (*p) ) {
 		case 32: goto st75;
 		case 111: goto st76;
@@ -1827,10 +1869,10 @@ case 134:
 	_out: {}
 	}
 
-#line 260 "src/admin.rl"
+#line 314 "src/admin.rl"
 
 
-	tbuf_ltrim(fiber->rbuf, (void *)pe - (void *)fiber->rbuf->data);
+	in->pos = pe;
 
 	if (p != pe) {
 		start(out);
@@ -1838,39 +1880,41 @@ case 134:
 		end(out);
 	}
 
-	return fiber_write(out->data, out->size);
+	coio_write(coio, out->data, out->size);
+	return 0;
 }
 
 static void
-admin_handler(void *data __attribute__((unused)))
+admin_handler(va_list ap)
 {
+	struct ev_io coio = va_arg(ap, struct ev_io);
+	struct iobuf *iobuf = va_arg(ap, struct iobuf *);
+	fiber_set_sid(fiber, box_sid());
 	lua_State *L = lua_newthread(tarantool_L);
 	int coro_ref = luaL_ref(tarantool_L, LUA_REGISTRYINDEX);
-	/** Allow to interrupt/kill administrative connections. */
-	fiber_setcancelstate(true);
 	@try {
 		for (;;) {
-			if (admin_dispatch(L) <= 0)
+			if (admin_dispatch(&coio, iobuf, L) < 0)
 				return;
+			iobuf_gc(iobuf);
 			fiber_gc();
 		}
 	} @finally {
 		luaL_unref(tarantool_L, LUA_REGISTRYINDEX, coro_ref);
+		evio_close(&coio);
+		iobuf_destroy(iobuf);
 	}
 }
 
-int
-admin_init(void)
+void
+admin_init(const char *bind_ipaddr, int admin_port)
 {
-	if (fiber_server("admin", cfg.admin_port, admin_handler, NULL, NULL) == NULL) {
-		say_syserror("can't bind to %d", cfg.admin_port);
-		return -1;
-	}
-	return 0;
+	static struct coio_service admin;
+	coio_service_init(&admin, "admin", bind_ipaddr,
+			  admin_port, admin_handler, NULL);
+	evio_service_start(&admin.evio_service);
 }
 
-
-
 /*
  * Local Variables:
  * mode: c
