@@ -1,29 +1,31 @@
 /*
- * Copyright (C) 2010 Mail.RU
- * Copyright (C) 2010 Yuriy Vostrikov
+ * Redistribution and use in source and binary forms, with or
+ * without modification, are permitted provided that the following
+ * conditions are met:
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
+ * 1. Redistributions of source code must retain the above
+ *    copyright notice, this list of conditions and the
+ *    following disclaimer.
  *
- * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * 2. Redistributions in binary form must reproduce the above
+ *    copyright notice, this list of conditions and the following
+ *    disclaimer in the documentation and/or other materials
+ *    provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
+ * <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
+ * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+ * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
+ * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  */
-
 #include "palloc.h"
 #include "third_party/valgrind/memcheck.h"
 #include "config.h"
@@ -38,8 +40,7 @@
 #include <stdlib.h>
 #include <third_party/queue.h>
 #include <tbuf.h>
-
-#define PALLOC_POOL_NAME_MAXLEN 16
+#include "exception.h"
 
 struct chunk {
 	uint32_t magic;
@@ -66,9 +67,9 @@ TAILQ_HEAD(class_tailq_head, chunk_class) classes;
 
 struct palloc_pool {
 	struct chunk_list_head chunks;
-	 SLIST_ENTRY(palloc_pool) link;
+	SLIST_ENTRY(palloc_pool) link;
 	size_t allocated;
-	const char *name;
+	char name[PALLOC_POOL_NAME_MAXLEN];
 };
 
 SLIST_HEAD(palloc_pool_head, palloc_pool) pools;
@@ -242,8 +243,10 @@ palloc_slow_path(struct palloc_pool *restrict pool, size_t size)
 {
 	struct chunk *chunk;
 	chunk = next_chunk_for(pool, size);
-	if (chunk == NULL)
-		abort();
+	if (chunk == NULL) {
+		tnt_raise(LoggedError, :ER_MEMORY_ISSUE,
+			  size, "palloc", "next chunk");
+	}
 
 	assert(chunk->free >= size);
 	void *ptr = chunk->brk;
@@ -319,8 +322,7 @@ ptruncate(struct palloc_pool *pool, size_t new_size)
 	ssize_t cut_size = pool->allocated - new_size;
 	struct chunk *chunk = SLIST_FIRST(&pool->chunks);
 
-	for (chunk = SLIST_FIRST(&pool->chunks); chunk;
-	     chunk = SLIST_FIRST(&pool->chunks)) {
+	for (; chunk; chunk = SLIST_FIRST(&pool->chunks)) {
 
 		size_t chunk_used = (chunk->size - chunk->free -
 				     sizeof(struct chunk));
@@ -355,6 +357,18 @@ prelease(struct palloc_pool *pool)
 	pool->allocated = 0;
 }
 
+void
+palloc_reset(struct palloc_pool *pool)
+{
+	struct chunk *chunk = SLIST_FIRST(&pool->chunks);
+	if (chunk) {
+		pool->allocated -= chunk->size - sizeof(struct chunk) -
+				   chunk->free;
+		chunk->free = chunk->size - sizeof(struct chunk);
+		chunk->brk = (void *)chunk + sizeof(struct chunk);
+	}
+}
+
 void
 prelease_after(struct palloc_pool *pool, size_t after)
 {
@@ -365,9 +379,12 @@ prelease_after(struct palloc_pool *pool, size_t after)
 struct palloc_pool *
 palloc_create_pool(const char *name)
 {
-	struct palloc_pool *pool = malloc(sizeof(struct palloc_pool));
-	assert(pool != NULL);
-	memset(pool, 0, sizeof(*pool));
+	struct palloc_pool *pool = calloc(sizeof(struct palloc_pool), 1);
+	if (pool == NULL) {
+		tnt_raise(LoggedError, :ER_MEMORY_ISSUE,
+			  sizeof(struct palloc_pool),
+			  "malloc", "palloc pool");
+	}
 	palloc_set_name(pool, name);
 	SLIST_INIT(&pool->chunks);
 	SLIST_INSERT_HEAD(&pools, pool, link);
@@ -456,7 +473,13 @@ palloc_stat(struct tbuf *buf)
 void
 palloc_set_name(struct palloc_pool *pool, const char *name)
 {
-	pool->name = name;
+	snprintf(pool->name, sizeof(pool->name), "%s", name);
+}
+
+const char *
+palloc_name(struct palloc_pool *pool)
+{
+	return pool->name;
 }
 
 size_t
