@@ -33,7 +33,7 @@
 #include "log_io.h"
 #include "fiber.h"
 #include "tarantool_pthread.h"
-#include "nio.h"
+#include "fio.h"
 #include "errinj.h"
 
 /*
@@ -102,46 +102,68 @@ struct recovery_state *recovery_state;
 
 static const u64 snapshot_cookie = 0;
 
+const char *wal_mode_STRS[] = { "none", "write", "fsync", "fsync_delay", NULL };
+
 /* {{{ LSN API */
 
 void
-wait_lsn_set(struct wait_lsn *wait_lsn, i64 lsn)
+wait_lsn_set(struct wait_lsn *wait_lsn, int64_t lsn)
 {
 	assert(wait_lsn->waiter == NULL);
 	wait_lsn->waiter = fiber;
 	wait_lsn->lsn = lsn;
 }
 
-int
-confirm_lsn(struct recovery_state *r, i64 lsn)
+
+/* Alert the waiter, if any. */
+static inline void
+wakeup_lsn_waiter(struct recovery_state *r)
+{
+	if (r->wait_lsn.waiter && r->confirmed_lsn >= r->wait_lsn.lsn)
+		fiber_wakeup(r->wait_lsn.waiter);
+}
+
+void
+confirm_lsn(struct recovery_state *r, int64_t lsn, bool is_commit)
 {
 	assert(r->confirmed_lsn <= r->lsn);
 
 	if (r->confirmed_lsn < lsn) {
-		if (r->confirmed_lsn + 1 != lsn)
-			say_warn("non consecutive lsn, last confirmed:%" PRIi64
-				 " new:%" PRIi64 " diff: %" PRIi64,
-				 r->confirmed_lsn, lsn, lsn - r->confirmed_lsn);
-		r->confirmed_lsn = lsn;
-		/* Alert the waiter, if any. There can be holes in
-		 * confirmed_lsn, in case of disk write failure,
-		 * but wal_writer never confirms LSNs out order.
-		 */
-		if (r->wait_lsn.waiter && r->confirmed_lsn >= r->wait_lsn.lsn) {
-			fiber_call(r->wait_lsn.waiter);
-		}
-
-		return 0;
+		if (is_commit) {
+			if (r->confirmed_lsn + 1 != lsn)
+				say_warn("non consecutive LSN, confirmed: %jd, "
+					 " new: %jd, diff: %jd",
+					 (intmax_t) r->confirmed_lsn,
+					 (intmax_t) lsn,
+					 (intmax_t) (lsn - r->confirmed_lsn));
+			r->confirmed_lsn = lsn;
+		 }
 	} else {
-		say_warn("lsn double confirmed:%" PRIi64, lsn);
+		 /*
+		 * There can be holes in
+		 * confirmed_lsn, in case of disk write failure, but
+		 * wal_writer never confirms LSNs out order.
+		 */
+		assert(false);
+		say_error("LSN is used twice or COMMIT order is broken: "
+			  "confirmed: %jd, new: %jd",
+			  (intmax_t) r->confirmed_lsn, (intmax_t) lsn);
 	}
+	wakeup_lsn_waiter(r);
+}
 
-	return -1;
+void
+set_lsn(struct recovery_state *r, int64_t lsn)
+{
+	r->lsn = lsn;
+	r->confirmed_lsn = lsn;
+	say_debug("set_lsn(%p, %" PRIi64, r, r->lsn);
+	wakeup_lsn_waiter(r);
 }
 
 /** Wait until the given LSN makes its way to disk. */
 void
-recovery_wait_lsn(struct recovery_state *r, i64 lsn)
+recovery_wait_lsn(struct recovery_state *r, int64_t lsn)
 {
 	while (lsn < r->confirmed_lsn) {
 		wait_lsn_set(&r->wait_lsn, lsn);
@@ -154,18 +176,15 @@ recovery_wait_lsn(struct recovery_state *r, i64 lsn)
 }
 
 
-i64
-next_lsn(struct recovery_state *r, i64 new_lsn)
+int64_t
+next_lsn(struct recovery_state *r)
 {
-	if (new_lsn == 0)
-		r->lsn++;
-	else
-		r->lsn = new_lsn;
-
-	say_debug("next_lsn(%p, %" PRIi64 ") => %" PRIi64, r, new_lsn, r->lsn);
+	r->lsn++;
+	say_debug("next_lsn(%p, %" PRIi64, r, r->lsn);
 	return r->lsn;
 }
 
+
 /* }}} */
 
 /* {{{ Initial recovery */
@@ -179,34 +198,36 @@ recovery_stop_local(struct recovery_state *r);
 
 void
 recovery_init(const char *snap_dirname, const char *wal_dirname,
-	      row_handler row_handler, int rows_per_wal,
-	      const char *wal_mode, double wal_fsync_delay, int flags)
+	      row_handler row_handler, void *row_handler_param,
+	      int rows_per_wal, int flags)
 {
 	assert(recovery_state == NULL);
 	recovery_state = p0alloc(eter_pool, sizeof(struct recovery_state));
 	struct recovery_state *r = recovery_state;
+	recovery_update_mode(r, "none", 0);
 
 	if (rows_per_wal <= 1)
 		panic("unacceptable value of 'rows_per_wal'");
 
 	r->row_handler = row_handler;
+	r->row_handler_param = row_handler_param;
 
 	r->snap_dir = &snap_dir;
 	r->snap_dir->dirname = strdup(snap_dirname);
 	r->wal_dir = &wal_dir;
 	r->wal_dir->dirname = strdup(wal_dirname);
-	r->wal_dir->open_wflags = strcasecmp(wal_mode, "fsync") ? 0 : WAL_SYNC_FLAG;
+	r->wal_dir->open_wflags = r->wal_mode == WAL_FSYNC ? WAL_SYNC_FLAG : 0;
 	r->rows_per_wal = rows_per_wal;
-	r->wal_fsync_delay = wal_fsync_delay;
 	wait_lsn_clear(&r->wait_lsn);
 	r->flags = flags;
 }
 
 void
-recovery_update_mode(const char *mode, double fsync_delay)
+recovery_update_mode(struct recovery_state *r,
+		     const char *mode, double fsync_delay)
 {
-	struct recovery_state *r = recovery_state;
-	(void) mode;
+	r->wal_mode = strindex(wal_mode_STRS, mode, WAL_MODE_MAX);
+	assert(r->wal_mode != WAL_MODE_MAX);
 	/* No mutex lock: let's not bother with whether
 	 * or not a WAL writer thread is present, and
 	 * if it's present, the delay will be propagated
@@ -217,9 +238,9 @@ recovery_update_mode(const char *mode, double fsync_delay)
 }
 
 void
-recovery_update_io_rate_limit(double new_limit)
+recovery_update_io_rate_limit(struct recovery_state *r, double new_limit)
 {
-	recovery_state->snap_io_rate_limit = new_limit * 1024 * 1024;
+	r->snap_io_rate_limit = new_limit * 1024 * 1024;
 }
 
 void
@@ -258,26 +279,27 @@ recovery_setup_panic(struct recovery_state *r, bool on_snap_error, bool on_wal_e
 
 /**
  * Read a snapshot and call row_handler for every snapshot row.
- *
- * @retval 0 success
- * @retval -1 failure
+ * Panic in case of error.
  */
-static int
+void
 recover_snap(struct recovery_state *r)
 {
-	int res = -1;
+	/*  current_wal isn't open during initial recover. */
+	assert(r->current_wal == NULL);
+	say_info("recovery start");
+
 	struct log_io *snap;
 	i64 lsn;
 
 	lsn = greatest_lsn(r->snap_dir);
 	if (lsn <= 0) {
 		say_error("can't find snapshot");
-		return -1;
+		goto error;
 	}
 	snap = log_io_open_for_read(r->snap_dir, lsn, NONE);
 	if (snap == NULL) {
 		say_error("can't find/open snapshot");
-		return -1;
+		goto error;
 	}
 	say_info("recover from `%s'", snap->filename);
 	struct log_io_cursor i;
@@ -286,17 +308,27 @@ recover_snap(struct recovery_state *r)
 
 	struct tbuf *row;
 	while ((row = log_io_cursor_next(&i))) {
-		if (r->row_handler(row) < 0) {
+		if (r->row_handler(r->row_handler_param, row) < 0) {
 			say_error("can't apply row");
-			goto end;
+			if (snap->dir->panic_if_error)
+				break;
 		}
 	}
-	r->lsn = r->confirmed_lsn = lsn;
-	res = 0;
-end:
 	log_io_cursor_close(&i);
 	log_io_close(&snap);
-	return res;
+
+	if (row == NULL) {
+		r->lsn = r->confirmed_lsn = lsn;
+		say_info("snapshot recovered, confirmed lsn: %"
+			 PRIi64, r->confirmed_lsn);
+		return;
+	}
+error:
+	if (greatest_lsn(r->snap_dir) <= 0) {
+		say_crit("didn't you forget to initialize storage with --init-storage switch?");
+		_exit(1);
+	}
+	panic("snapshot recovery failed");
 }
 
 #define LOG_EOF 0
@@ -325,12 +357,12 @@ recover_wal(struct recovery_state *r, struct log_io *l)
 		 * After handler(row) returned, row may be
 		 * modified, do not use it.
 		 */
-		if (r->row_handler(row) < 0) {
+		if (r->row_handler(r->row_handler_param, row) < 0) {
 			say_error("can't apply row");
-			goto end;
+			if (l->dir->panic_if_error)
+				goto end;
 		}
-		next_lsn(r, lsn);
-		confirm_lsn(r, lsn);
+		set_lsn(r, lsn);
 	}
 	res = i.eof_read ? LOG_EOF : 1;
 end:
@@ -443,7 +475,6 @@ recover_current_wal:
 				  r->current_wal->filename);
 			break;
 		}
-
 		if (result == LOG_EOF) {
 			say_info("done `%s' confirmed_lsn: %" PRIi64,
 				 r->current_wal->filename,
@@ -465,48 +496,16 @@ recover_current_wal:
 	return result;
 }
 
+/**
+ * Recover all WALs created after the last snapshot. Panic if
+ * error.
+ */
 void
-recover(struct recovery_state *r, i64 lsn)
+recover_existing_wals(struct recovery_state *r)
 {
-	/* * current_wal isn't open during initial recover. */
-	assert(r->current_wal == NULL);
-	/*
-	 * If the caller sets confirmed_lsn to a non-zero value,
-	 * snapshot recovery is skipped and we proceed directly to
-	 * finding the WAL with the respective LSN and continue
-	 * recovery from this WAL.  @fixme: this is a gotcha, due
-	 * to whihc a replica is unable to read data from a master
-	 * if the replica has no snapshot or the master has no WAL
-	 * with the requested LSN.
-	 */
-	say_info("recovery start");
-	if (lsn == 0) {
-		if (recover_snap(r) != 0) {
-			if (greatest_lsn(r->snap_dir) <= 0) {
-				say_crit("didn't you forget to initialize storage with --init-storage switch?");
-				_exit(1);
-			}
-			panic("snapshot recovery failed");
-		}
-		say_info("snapshot recovered, confirmed lsn: %"
-			 PRIi64, r->confirmed_lsn);
-	} else {
-		/*
-		 * Note that recovery starts with lsn _NEXT_ to
-		 * the confirmed one.
-		 */
-		r->lsn = r->confirmed_lsn = lsn - 1;
-	}
 	i64 next_lsn = r->confirmed_lsn + 1;
 	i64 wal_lsn = find_including_file(r->wal_dir, next_lsn);
 	if (wal_lsn <= 0) {
-		if (lsn != 0) {
-			/*
-			 * Recovery for replication relay, did not
-			 * find the requested LSN.
-			 */
-			say_error("can't find WAL containing record with lsn: %" PRIi64, next_lsn);
-		}
 		/* No WALs to recover from. */
 		goto out;
 	}
@@ -550,7 +549,7 @@ recovery_finalize(struct recovery_state *r)
 		} else if (r->current_wal->rows == 1) {
 			/* Rename inprogress wal with one row */
 			say_warn("rename unfinished %s WAL", r->current_wal->filename);
-			if (inprogress_log_rename(r->current_wal->filename) != 0)
+			if (inprogress_log_rename(r->current_wal) != 0)
 				panic("can't rename 'inprogress' WAL");
 		} else
 			panic("too many rows in inprogress WAL `%s'", r->current_wal->filename);
@@ -693,16 +692,15 @@ STAILQ_HEAD(wal_fifo, wal_write_request);
 
 struct wal_writer
 {
-	struct wal_fifo input_queue;
-	struct wal_fifo commit_queue;
-	struct wal_fifo rollback_queue;
+	struct wal_fifo input;
+	struct wal_fifo commit;
 	pthread_t thread;
 	pthread_mutex_t mutex;
 	pthread_cond_t cond;
-	ev_async commit_event;
-	ev_async rollback_event;
-	struct nbatch *batch;
+	ev_async write_event;
+	struct fio_batch *batch;
 	bool is_shutdown;
+	bool is_rollback;
 };
 
 static pthread_once_t wal_writer_once = PTHREAD_ONCE_INIT;
@@ -743,14 +741,13 @@ wal_writer_init_once()
 
 /**
  * A commit watcher callback is invoked whenever there
- * are requests in wal_writer->commit_queue. This callback is
+ * are requests in wal_writer->commit. This callback is
  * associated with an internal WAL writer watcher and is
  * invoked in the front-end main event loop.
  *
  * A rollback watcher callback is invoked only when there is
- * no input in writer->input_queue (we processed all pending
- * requests, and either put them to commit_queue or to
- * rollback_queue).
+ * a rollback request and commit is empty.
+ * We roll back the entire input queue.
  *
  * ev_async, under the hood, is a simple pipe. The WAL
  * writer thread writes to that pipe whenever it's done
@@ -758,37 +755,35 @@ wal_writer_init_once()
  * call in the writer thread loop).
  */
 static void
-wal_schedule_queue(struct wal_writer *writer, struct wal_fifo *qptr)
+wal_schedule_queue(struct wal_fifo *queue)
 {
-	struct wal_fifo queue = STAILQ_HEAD_INITIALIZER(queue);
-
-	(void) tt_pthread_mutex_lock(&writer->mutex);
-	STAILQ_CONCAT(&queue, qptr);
-	(void) tt_pthread_mutex_unlock(&writer->mutex);
 	/*
 	 * Can't use STAILQ_FOREACH since fiber_call()
 	 * destroys the list entry.
 	 */
 	struct wal_write_request *req, *tmp;
-	STAILQ_FOREACH_SAFE(req, &queue, wal_fifo_entry, tmp)
+	STAILQ_FOREACH_SAFE(req, queue, wal_fifo_entry, tmp)
 		fiber_call(req->fiber);
 }
 
 static void
-wal_schedule_commit(ev_watcher *watcher, int event __attribute__((unused)))
+wal_schedule(ev_watcher *watcher, int event __attribute__((unused)))
 {
 	struct wal_writer *writer = watcher->data;
-	wal_schedule_queue(writer, &writer->commit_queue);
-}
+	struct wal_fifo commit = STAILQ_HEAD_INITIALIZER(commit);
+	struct wal_fifo rollback = STAILQ_HEAD_INITIALIZER(rollback);
 
-static void
-wal_schedule_rollback(ev_watcher *watcher, int event __attribute__((unused)))
-{
-	struct wal_writer *writer = watcher->data;
-	/* First, commit pending commit transactions, if any. */
-	wal_schedule_queue(writer, &writer->commit_queue);
-	/* Rollback queue is already in reverse order, thanks to wal_writer */
-	wal_schedule_queue(writer, &writer->rollback_queue);
+	(void) tt_pthread_mutex_lock(&writer->mutex);
+	STAILQ_CONCAT(&commit, &writer->commit);
+	if (writer->is_rollback) {
+		STAILQ_CONCAT(&rollback, &writer->input);
+		writer->is_rollback = false;
+	}
+	(void) tt_pthread_mutex_unlock(&writer->mutex);
+
+	wal_schedule_queue(&commit);
+	STAILQ_REVERSE(&rollback, wal_write_request, wal_fifo_entry);
+	wal_schedule_queue(&rollback);
 }
 
 /**
@@ -813,21 +808,18 @@ wal_writer_init(struct wal_writer *writer)
 
 	(void) tt_pthread_cond_init(&writer->cond, NULL);
 
-	STAILQ_INIT(&writer->input_queue);
-	STAILQ_INIT(&writer->commit_queue);
-	STAILQ_INIT(&writer->rollback_queue);
+	STAILQ_INIT(&writer->input);
+	STAILQ_INIT(&writer->commit);
 
-	ev_async_init(&writer->commit_event, (void *)wal_schedule_commit);
-	writer->commit_event.data = writer;
-	ev_async_init(&writer->rollback_event, (void *)wal_schedule_rollback);
-	writer->rollback_event.data = writer;
+	ev_async_init(&writer->write_event, (void *)wal_schedule);
+	writer->write_event.data = writer;
 
 	(void) tt_pthread_once(&wal_writer_once, wal_writer_init_once);
 
-	writer->batch = nbatch_alloc(sysconf(_SC_IOV_MAX));
+	writer->batch = fio_batch_alloc(sysconf(_SC_IOV_MAX));
 
 	if (writer->batch == NULL)
-		panic_syserror("nbatch_alloc");
+		panic_syserror("fio_batch_alloc");
 }
 
 /** Destroy a WAL writer structure. */
@@ -860,17 +852,15 @@ wal_writer_start(struct recovery_state *r)
 	assert(r->writer == NULL);
 	assert(r->watcher == NULL);
 	assert(r->current_wal == NULL);
-	assert(wal_writer.is_shutdown == false);
-	assert(STAILQ_EMPTY(&wal_writer.input_queue));
-	assert(STAILQ_EMPTY(&wal_writer.commit_queue));
-	assert(STAILQ_EMPTY(&wal_writer.rollback_queue));
+	assert(! wal_writer.is_shutdown);
+	assert(STAILQ_EMPTY(&wal_writer.input));
+	assert(STAILQ_EMPTY(&wal_writer.commit));
 
 	/* I. Initialize the state. */
 	wal_writer_init(&wal_writer);
 	r->writer = &wal_writer;
 
-	ev_async_start(&wal_writer.commit_event);
-	ev_async_start(&wal_writer.rollback_event);
+	ev_async_start(&wal_writer.write_event);
 
 	/* II. Start the thread. */
 
@@ -900,8 +890,7 @@ wal_writer_stop(struct recovery_state *r)
 		panic_syserror("WAL writer: thread join failed");
 	}
 
-	ev_async_stop(&writer->commit_event);
-	ev_async_stop(&writer->rollback_event);
+	ev_async_stop(&writer->write_event);
 	wal_writer_destroy(writer);
 
 	r->writer = NULL;
@@ -913,14 +902,16 @@ wal_writer_stop(struct recovery_state *r)
  * do. Loop in case of a spurious wakeup.
  */
 void
-wal_writer_pop(struct wal_writer *writer, struct wal_fifo *input, bool wait)
+wal_writer_pop(struct wal_writer *writer, struct wal_fifo *input)
 {
-	do {
-		STAILQ_CONCAT(input, &writer->input_queue);
-		if (STAILQ_EMPTY(input) == false || wait == false)
+	while (! writer->is_shutdown)
+	{
+		if (! writer->is_rollback && ! STAILQ_EMPTY(&writer->input)) {
+			STAILQ_CONCAT(input, &writer->input);
 			break;
+		}
 		(void) tt_pthread_cond_wait(&writer->cond, &writer->mutex);
-	} while (writer->is_shutdown == false);
+	}
 }
 
 /**
@@ -976,7 +967,7 @@ wal_opt_rotate(struct log_io **wal, int rows_per_wal, struct log_dir *dir, u64 l
 		 * Rename WAL after the first successful write
 		 * to a name  without .inprogress suffix.
 		 */
-		if (inprogress_log_rename(l->filename))
+		if (inprogress_log_rename(l))
 			log_io_close(&l);       /* error. */
 	}
 	assert(wal_to_close == NULL);
@@ -1001,27 +992,27 @@ wal_opt_sync(struct log_io *wal, double sync_delay)
 }
 
 static struct wal_write_request *
-wal_fill_batch(struct log_io *wal, struct nbatch *batch, int rows_per_wal,
+wal_fill_batch(struct log_io *wal, struct fio_batch *batch, int rows_per_wal,
 	       struct wal_write_request *req)
 {
 	int max_rows = wal->is_inprogress ? 1 : rows_per_wal - wal->rows;
 	/* Post-condition of successful by wal_opt_rotate(). */
 	assert(max_rows > 0);
-	nbatch_start(batch, max_rows);
-	while (req != NULL && nbatch_is_full(batch) == false) {
+	fio_batch_start(batch, max_rows);
+	while (req != NULL && ! fio_batch_is_full(batch)) {
 		struct row_v11 *row = &req->row;
 		header_v11_sign(&row->header);
-		nbatch_add(batch, row, row_v11_size(row));
+		fio_batch_add(batch, row, row_v11_size(row));
 		req = STAILQ_NEXT(req, wal_fifo_entry);
 	}
 	return req;
 }
 
 static struct wal_write_request *
-wal_write_batch(struct log_io *wal, struct nbatch *batch,
+wal_write_batch(struct log_io *wal, struct fio_batch *batch,
 		struct wal_write_request *req, struct wal_write_request *end)
 {
-	int rows_written = nbatch_write(batch, fileno(wal->f));
+	int rows_written = fio_batch_write(batch, fileno(wal->f));
 	wal->rows += rows_written;
 	while (req != end && rows_written-- != 0)  {
 		req->res = 0;
@@ -1032,13 +1023,13 @@ wal_write_batch(struct log_io *wal, struct nbatch *batch,
 
 static void
 wal_write_to_disk(struct recovery_state *r, struct wal_writer *writer,
-		  struct wal_fifo *input_queue, struct wal_fifo *commit_queue,
-		  struct wal_fifo *rollback_queue)
+		  struct wal_fifo *input, struct wal_fifo *commit,
+		  struct wal_fifo *rollback)
 {
 	struct log_io **wal = &r->current_wal;
-	struct nbatch *batch = writer->batch;
+	struct fio_batch *batch = writer->batch;
 
-	struct wal_write_request *req = STAILQ_FIRST(input_queue);
+	struct wal_write_request *req = STAILQ_FIRST(input);
 	struct wal_write_request *write_end = req;
 
 	while (req) {
@@ -1053,8 +1044,8 @@ wal_write_to_disk(struct recovery_state *r, struct wal_writer *writer,
 		wal_opt_sync(*wal, r->wal_fsync_delay);
 		req = write_end;
 	}
-	STAILQ_SPLICE(input_queue, req, wal_fifo_entry, rollback_queue);
-	STAILQ_CONCAT(commit_queue, input_queue);
+	STAILQ_SPLICE(input, write_end, wal_fifo_entry, rollback);
+	STAILQ_CONCAT(commit, input);
 }
 
 /** WAL writer thread main loop.  */
@@ -1063,48 +1054,31 @@ wal_writer_thread(void *worker_args)
 {
 	struct recovery_state *r = worker_args;
 	struct wal_writer *writer = r->writer;
-	bool input_was_empty = true;
-	struct wal_fifo input_queue = STAILQ_HEAD_INITIALIZER(input_queue);
-	struct wal_fifo commit_queue = STAILQ_HEAD_INITIALIZER(commit_queue);
-	struct wal_fifo rollback_queue = STAILQ_HEAD_INITIALIZER(rollback_queue);
+	struct wal_fifo input = STAILQ_HEAD_INITIALIZER(input);
+	struct wal_fifo commit = STAILQ_HEAD_INITIALIZER(commit);
+	struct wal_fifo rollback = STAILQ_HEAD_INITIALIZER(rollback);
 
 	(void) tt_pthread_mutex_lock(&writer->mutex);
-	while (writer->is_shutdown == false) {
-		bool in_rollback = ! STAILQ_EMPTY(&writer->rollback_queue);
-		wal_writer_pop(writer, &input_queue, input_was_empty && !in_rollback);
+	while (! writer->is_shutdown) {
+		wal_writer_pop(writer, &input);
 		(void) tt_pthread_mutex_unlock(&writer->mutex);
-		if (input_was_empty == false) {
-			/*
-			 * Wake up fibers waiting on the old list *here*
-			 * since we need a membar for request res to
-			 * sync up.
-			 */
-			ev_async_send(&writer->commit_event);
-		} else if (in_rollback && STAILQ_EMPTY(&input_queue)) {
-			/*
-			 * We can initiate a rollback only when
-			 * all input has been put to the rollback
-			 * queue, since rollback must start from
-			 * the last processed request.
-			 */
-			ev_async_send(&writer->rollback_event);
-		}
-		input_was_empty = STAILQ_EMPTY(&input_queue);
-		if (in_rollback) {
-			STAILQ_CONCAT(&rollback_queue, &input_queue);
-		} else {
-			wal_write_to_disk(r, writer, &input_queue,
-					  &commit_queue, &rollback_queue);
-		}
+
+		wal_write_to_disk(r, writer, &input, &commit, &rollback);
+
 		(void) tt_pthread_mutex_lock(&writer->mutex);
-		STAILQ_CONCAT(&writer->commit_queue, &commit_queue);
-		/* Rollback is done in reverse order. */
-		if (STAILQ_EMPTY(&rollback_queue) == false) {
-			STAILQ_REVERSE(&rollback_queue, wal_write_request,
-				       wal_fifo_entry);
-			STAILQ_CONCAT(&rollback_queue, &writer->rollback_queue);
-			STAILQ_CONCAT(&writer->rollback_queue, &rollback_queue);
+		STAILQ_CONCAT(&writer->commit, &commit);
+		if (! STAILQ_EMPTY(&rollback)) {
+			/*
+			 * Begin rollback: create a rollback queue
+			 * from all requests which were not
+			 * written to disk and all requests in the
+			 * input queue.
+			 */
+			writer->is_rollback = true;
+			STAILQ_CONCAT(&rollback, &writer->input);
+			STAILQ_CONCAT(&writer->input, &rollback);
 		}
+		ev_async_send(&writer->write_event);
 	}
 	(void) tt_pthread_mutex_unlock(&writer->mutex);
 	if (r->current_wal != NULL)
@@ -1123,6 +1097,9 @@ wal_write(struct recovery_state *r, i64 lsn, u64 cookie,
 	say_debug("wal_write lsn=%" PRIi64, lsn);
 	ERROR_INJECT_RETURN(ERRINJ_WAL_IO);
 
+	if (r->wal_mode == WAL_NONE)
+		return 0;
+
 	struct wal_writer *writer = r->writer;
 
 	struct wal_write_request *req =
@@ -1136,23 +1113,15 @@ wal_write(struct recovery_state *r, i64 lsn, u64 cookie,
 
 	(void) tt_pthread_mutex_lock(&writer->mutex);
 
-	bool input_was_empty = STAILQ_EMPTY(&writer->input_queue);
-	bool in_rollback = ! STAILQ_EMPTY(&writer->rollback_queue);
-	/*
-	 * If the rollback queue is not empty,
-	 * abort the request right away until we've finished
-	 * with rollback.
-	 */
-	if (! in_rollback) {
-		STAILQ_INSERT_TAIL(&writer->input_queue, req, wal_fifo_entry);
-		if (input_was_empty)
-			(void) tt_pthread_cond_signal(&writer->cond);
-	}
+	bool input_was_empty = STAILQ_EMPTY(&writer->input);
+	STAILQ_INSERT_TAIL(&writer->input, req, wal_fifo_entry);
+
+	if (input_was_empty)
+		(void) tt_pthread_cond_signal(&writer->cond);
 
 	(void) tt_pthread_mutex_unlock(&writer->mutex);
 
-	if (! in_rollback)
-		fiber_yield(); /* Request was inserted. */
+	fiber_yield(); /* Request was inserted. */
 
 	return req->res;
 }
@@ -1161,8 +1130,19 @@ wal_write(struct recovery_state *r, i64 lsn, u64 cookie,
 
 /* {{{ SAVE SNAPSHOT and tarantool_box --cat */
 
+static void
+snap_write_batch(struct fio_batch *batch, int fd)
+{
+	int rows_written = fio_batch_write(batch, fd);
+	if (rows_written != batch->rows) {
+		say_error("partial write: %d out of %d rows",
+			  rows_written, batch->rows);
+		panic_syserror("fio_batch_write");
+	}
+}
+
 void
-snapshot_write_row(struct log_io *l, struct nbatch *batch,
+snapshot_write_row(struct log_io *l, struct fio_batch *batch,
 		   const void *metadata, size_t metadata_len,
 		   const void *data, size_t data_len)
 {
@@ -1179,15 +1159,14 @@ snapshot_write_row(struct log_io *l, struct nbatch *batch,
 		     metadata, metadata_len, data, data_len);
 	header_v11_sign(&row->header);
 
-	nbatch_add(batch, row, row_v11_size(row));
+	fio_batch_add(batch, row, row_v11_size(row));
 
 	if (++rows % 100000 == 0)
 		say_crit("%.1fM rows written", rows / 1000000.);
 
-	if (nbatch_is_full(batch)) {
-		if (nbatch_write(batch, fileno(l->f)) != batch->rows)
-			panic_syserror("nbatch_write");
-		nbatch_start(batch, INT_MAX);
+	if (fio_batch_is_full(batch)) {
+		snap_write_batch(batch, fileno(l->f));
+		fio_batch_start(batch, INT_MAX);
 		prelease_after(fiber->gc_pool, 128 * 1024);
 	}
 
@@ -1213,38 +1192,29 @@ snapshot_write_row(struct log_io *l, struct nbatch *batch,
 
 void
 snapshot_save(struct recovery_state *r,
-	      void (*f) (struct log_io *, struct nbatch *))
+	      void (*f) (struct log_io *, struct fio_batch *))
 {
 	struct log_io *snap;
 	snap = log_io_open_for_write(r->snap_dir, r->confirmed_lsn,
 				     INPROGRESS);
 	if (snap == NULL)
 		panic_status(errno, "Failed to save snapshot: failed to open file in write mode.");
-	struct nbatch *batch = nbatch_alloc(sysconf(_SC_IOV_MAX));
+	struct fio_batch *batch = fio_batch_alloc(sysconf(_SC_IOV_MAX));
 	if (batch == NULL)
-		panic_syserror("malloc");
-	nbatch_start(batch, INT_MAX);
+		panic_syserror("fio_batch_alloc");
+	fio_batch_start(batch, INT_MAX);
 	/*
 	 * While saving a snapshot, snapshot name is set to
 	 * <lsn>.snap.inprogress. When done, the snapshot is
 	 * renamed to <lsn>.snap.
 	 */
-	const char *final_filename =
-		format_filename(r->snap_dir, r->confirmed_lsn, NONE);
-	say_info("saving snapshot `%s'", final_filename);
+	say_info("saving snapshot `%s'",
+		 format_filename(r->snap_dir, r->confirmed_lsn,
+				 NONE));
 	f(snap, batch);
 
-	if (batch->rows && nbatch_write(batch, fileno(snap->f)) != batch->rows)
-		panic_syserror("nbatch_write");
-
-	if (log_io_sync(snap) < 0)
-		panic("fsync");
-
-	if (link(snap->filename, final_filename) == -1)
-		panic_status(errno, "can't create hard link to snapshot");
-
-	if (unlink(snap->filename) == -1)
-		say_syserror("can't unlink 'inprogress' snapshot");
+	if (batch->rows)
+		snap_write_batch(batch, fileno(snap->f));
 
 	free(batch);
 	log_io_close(&snap);
@@ -1261,7 +1231,8 @@ snapshot_save(struct recovery_state *r,
 
 int
 read_log(const char *filename,
-	 row_handler *xlog_handler, row_handler *snap_handler)
+	 row_handler *xlog_handler, row_handler *snap_handler,
+	 void *param)
 {
 	struct log_dir *dir;
 	row_handler *h;
@@ -1284,7 +1255,7 @@ read_log(const char *filename,
 	log_io_cursor_open(&i, l);
 	struct tbuf *row;
 	while ((row = log_io_cursor_next(&i)))
-		h(row);
+		h(param, row);
 
 	log_io_cursor_close(&i);
 	log_io_close(&l);
