@@ -1,26 +1,29 @@
 /*
- * Copyright (C) 2010-2011 Mail.RU
- * Copyright (C) 2010-2011 Yuriy Vostrikov
+ * Redistribution and use in source and binary forms, with or
+ * without modification, are permitted provided that the following
+ * conditions are met:
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
+ * 1. Redistributions of source code must retain the above
+ *    copyright notice, this list of conditions and the
+ *    following disclaimer.
  *
- * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * 2. Redistributions in binary form must reproduce the above
+ *    copyright notice, this list of conditions and the following
+ *    disclaimer in the documentation and/or other materials
+ *    provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY <COPYRIGHT HOLDER> ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
+ * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
+ * <COPYRIGHT HOLDER> OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
+ * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
+ * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
+ * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  */
 #include "tarantool.h"
@@ -45,6 +48,7 @@
 #include <admin.h>
 #include <replication.h>
 #include <fiber.h>
+#include <coeio.h>
 #include <iproto.h>
 #include <latch.h>
 #include <recovery.h>
@@ -59,7 +63,10 @@
 #include <third_party/gopt/gopt.h>
 #include <cfg/warning.h>
 #include "tarantool_pthread.h"
-#include "tarantool_lua.h"
+#include "lua/init.h"
+#include "memcached.h"
+#include "session.h"
+#include "box/box.h"
 
 
 static pid_t master_pid;
@@ -75,25 +82,50 @@ static void *main_opt = NULL;
 struct tarantool_cfg cfg;
 static ev_signal *sigs = NULL;
 
+int snapshot_pid = 0; /* snapshot processes pid */
 bool init_storage, booting = true;
 
 static int
 core_check_config(struct tarantool_cfg *conf)
 {
-	/* Check that the mode is a supported one. */
-	if (strcmp(conf->wal_mode, "fsync") != 0 &&
-	    strcmp(conf->wal_mode, "fsync_delay") != 0) {
-		out_warning(0, "wal_mode is not one of 'fsync', 'fsync_delay'");
+	if (strindex(wal_mode_STRS, conf->wal_mode,
+		     WAL_MODE_MAX) == WAL_MODE_MAX) {
+		out_warning(0, "wal_mode %s is not recognized", conf->wal_mode);
 		return -1;
 	}
 	return 0;
 }
 
+void
+title(const char *fmt, ...)
+{
+	va_list ap;
+	char buf[128], *bufptr = buf, *bufend = buf + sizeof(buf);
+
+	va_start(ap, fmt);
+	bufptr += vsnprintf(bufptr, bufend - bufptr, fmt, ap);
+	va_end(ap);
+
+	int ports[] = { cfg.primary_port, cfg.secondary_port,
+			cfg.memcached_port, cfg.admin_port,
+			cfg.replication_port };
+	int *pptr = ports;
+	char *names[] = { "pri", "sec", "memc", "adm", "rpl", NULL };
+	char **nptr = names;
+
+	for (; *nptr; nptr++, pptr++)
+		if (*pptr)
+			bufptr += snprintf(bufptr, bufend - bufptr,
+					   " %s: %i", *nptr, *pptr);
+
+	set_proc_title(buf);
+}
+
 static i32
 load_cfg(struct tarantool_cfg *conf, i32 check_rdonly)
 {
 	FILE *f;
-	i32 n_accepted, n_skipped;
+	i32 n_accepted, n_skipped, n_ignored;
 
 	tbuf_reset(cfg_out);
 
@@ -104,26 +136,36 @@ load_cfg(struct tarantool_cfg *conf, i32 check_rdonly)
 
 	if (f == NULL) {
 		out_warning(0, "can't open config `%s'", cfg_filename);
-
 		return -1;
 	}
 
-	parse_cfg_file_tarantool_cfg(conf, f, check_rdonly, &n_accepted, &n_skipped);
-
+	int syntax = parse_cfg_file_tarantool_cfg(conf, f, check_rdonly,
+						  &n_accepted,
+						  &n_skipped,
+						  &n_ignored);
 	fclose(f);
+
+	if (syntax != 0)
+		return -1;
+
 	if (check_cfg_tarantool_cfg(conf) != 0)
 		return -1;
 
-	if (n_accepted == 0 || n_skipped != 0)
+	if (n_skipped != 0)
 		return -1;
 
+	if (n_accepted == 0) {
+		out_warning(0, "empty configuration file '%s'", cfg_filename);
+		return -1;
+	}
+
 	if (core_check_config(conf) != 0)
 		return -1;
 
 	if (replication_check_config(conf) != 0)
 		return -1;
 
-	return mod_check_config(conf);
+	return box_check_config(conf);
 }
 
 static int
@@ -157,9 +199,9 @@ core_reload_config(const struct tarantool_cfg *old_conf,
 		say_debug("%s: wal_fsync_delay [%f] -> [%f]",
 			__func__, old_conf->wal_fsync_delay, new_delay);
 
-	recovery_update_mode(new_conf->wal_mode, new_delay);
+	recovery_update_mode(recovery_state, new_conf->wal_mode, new_delay);
 
-	recovery_update_io_rate_limit(new_conf->snap_io_rate_limit);
+	recovery_update_io_rate_limit(recovery_state, new_conf->snap_io_rate_limit);
 
 	ev_set_io_collect_interval(new_conf->io_collect_interval);
 
@@ -174,7 +216,7 @@ reload_cfg(struct tbuf *out)
 
 	if (latch == NULL) {
 		latch = palloc(eter_pool, sizeof(*latch));
-		tnt_latch_init(latch);
+		tnt_latch_create(latch);
 	}
 
 	if (tnt_latch_trylock(latch) == -1) {
@@ -221,7 +263,7 @@ reload_cfg(struct tbuf *out)
 			return -1;
 
 		/* Now pass the config to the module, to take action. */
-		if (mod_reload_config(&cfg, &new_cfg) != 0)
+		if (box_reload_config(&cfg, &new_cfg) != 0)
 			return -1;
 		/* All OK, activate the config. */
 		swap_tarantool_cfg(&cfg, &new_cfg);
@@ -240,10 +282,29 @@ reload_cfg(struct tbuf *out)
 	return 0;
 }
 
+/** Print the configuration file in YAML format. */
+void
+show_cfg(struct tbuf *out)
+{
+	tarantool_cfg_iterator_t *i;
+	char *key, *value;
+
+	tbuf_printf(out, "configuration:" CRLF);
+	i = tarantool_cfg_iterator_init();
+	while ((key = tarantool_cfg_iterator_next(i, &cfg, &value)) != NULL) {
+		if (value) {
+			tbuf_printf(out, "  %s: \"%s\"" CRLF, key, value);
+			free(value);
+		} else {
+			tbuf_printf(out, "  %s: (null)" CRLF, key);
+		}
+	}
+}
+
 const char *
 tarantool_version(void)
 {
-	return TARANTOOL_VERSION;
+	return PACKAGE_VERSION;
 }
 
 static double start_time;
@@ -257,6 +318,9 @@ tarantool_uptime(void)
 int
 snapshot(void *ev, int events __attribute__((unused)))
 {
+	if (snapshot_pid)
+		return EINPROGRESS;
+
 	pid_t p = fork();
 	if (p < 0) {
 		say_syserror("fork");
@@ -278,10 +342,11 @@ snapshot(void *ev, int events __attribute__((unused)))
 		 * make 'save snapshot' synchronous, and propagate
 		 * any possible error up to the user.
 		 */
-		wait_for_child(p);
-		assert(p == fiber->cw.rpid);
-		return (WIFSIGNALED(fiber->cw.rstatus) ? EINTR :
-			WEXITSTATUS(fiber->cw.rstatus));
+
+		snapshot_pid = p;
+		int status = wait_for_child(p);
+		snapshot_pid = 0;
+		return (WIFSIGNALED(status) ? EINTR : WEXITSTATUS(status));
 	}
 
 	fiber_set_name(fiber, "dumper");
@@ -292,7 +357,7 @@ snapshot(void *ev, int events __attribute__((unused)))
 	 * parent stdio buffers at exit().
 	 */
 	close_all_xcpt(1, sayfd);
-	snapshot_save(recovery_state, mod_snapshot);
+	snapshot_save(recovery_state, box_snapshot);
 
 	exit(EXIT_SUCCESS);
 	return 0;
@@ -301,10 +366,64 @@ snapshot(void *ev, int events __attribute__((unused)))
 static void
 signal_cb(void)
 {
-	/* terminating main event loop */
+	/* Terminate the main event loop */
 	ev_unloop(EV_A_ EVUNLOOP_ALL);
 }
 
+/** Try to log as much as possible before dumping a core.
+ *
+ * Core files are not aways allowed and it takes an effort to
+ * extract useful information from them.
+ *
+ * *Recursive invocation*
+ *
+ * Unless SIGSEGV is sent by kill(), Linux
+ * resets the signal a default value before invoking
+ * the handler.
+ *
+ * Despite that, as an extra precaution to avoid infinite
+ * recursion, we count invocations of the handler, and
+ * quietly _exit() when called for a second time.
+ */
+static void
+sig_fatal_cb(int signo)
+{
+	static volatile sig_atomic_t in_cb = 0;
+	int fd = STDERR_FILENO;
+	struct sigaction sa;
+
+	/* Got a signal while running the handler. */
+	if (in_cb) {
+		fdprintf(fd, "Fatal %d while backtracing", signo);
+		goto end;
+	}
+
+	in_cb = 1;
+
+	if (signo == SIGSEGV)
+		fdprintf(fd, "Segmentation fault\n");
+	else
+		fdprintf(fd, "Got a fatal signal %d\n", signo);
+
+	fdprintf(fd, "Current time: %u\n", (unsigned) time(0));
+	fdprintf(fd,
+		 "Please file a bug at http://bugs.launchpad.net/tarantool\n"
+		 "Attempting backtrace... Note: since the server has "
+		 "already crashed, \nthis may fail as well\n");
+#ifdef ENABLE_BACKTRACE
+	print_backtrace();
+#endif
+end:
+	/* Try to dump core. */
+	memset(&sa, 0, sizeof(sa));
+	sigemptyset(&sa.sa_mask);
+	sa.sa_handler = SIG_DFL;
+	sigaction(SIGABRT, &sa, NULL);
+
+	abort();
+}
+
+
 static void
 signal_free(void)
 {
@@ -330,7 +449,9 @@ signal_reset()
 	if (sigaction(SIGUSR1, &sa, NULL) == -1 ||
 	    sigaction(SIGINT, &sa, NULL) == -1 ||
 	    sigaction(SIGTERM, &sa, NULL) == -1 ||
-	    sigaction(SIGHUP, &sa, NULL) == -1)
+	    sigaction(SIGHUP, &sa, NULL) == -1 ||
+	    sigaction(SIGSEGV, &sa, NULL) == -1 ||
+	    sigaction(SIGFPE, &sa, NULL) == -1)
 		say_syserror("sigaction");
 
 	/* Unblock any signals blocked by libev. */
@@ -358,6 +479,14 @@ signal_init(void)
 		exit(EX_OSERR);
 	}
 
+	sa.sa_handler = sig_fatal_cb;
+
+	if (sigaction(SIGSEGV, &sa, 0) == -1 ||
+	    sigaction(SIGFPE, &sa, 0) == -1) {
+		say_syserror("sigaction");
+		exit(EX_OSERR);
+	}
+
 	sigs = palloc(eter_pool, sizeof(ev_signal) * 4);
 	memset(sigs, 0, sizeof(ev_signal) * 4);
 	ev_signal_init(&sigs[0], (void*)snapshot, SIGUSR1);
@@ -438,14 +567,14 @@ background()
 	close(STDERR_FILENO);
 	return;
 error:
-        exit(EXIT_FAILURE);
+	exit(EXIT_FAILURE);
 }
 
 void
 tarantool_free(void)
 {
 	/*
-	 * Got to be done prior to anything else, since GC 
+	 * Got to be done prior to anything else, since GC
 	 * handlers can refer to other subsystems (e.g. fibers).
 	 */
 	if (tarantool_L)
@@ -465,6 +594,7 @@ tarantool_free(void)
 		unlink(cfg.pid_file);
 	destroy_tarantool_cfg(&cfg);
 
+	session_free();
 	fiber_free();
 	palloc_free();
 	ev_default_destroy();
@@ -482,6 +612,7 @@ initialize(double slab_alloc_arena, int slab_alloc_minimal, double slab_alloc_fa
 	if (!salloc_init(slab_alloc_arena * (1 << 30), slab_alloc_minimal, slab_alloc_factor))
 		panic_syserror("can't initialize slab allocator");
 	fiber_init();
+	coeio_init();
 }
 
 static void
@@ -549,11 +680,11 @@ main(int argc, char **argv)
 	binary_filename = argv[0];
 
 	if (gopt(opt, 'V')) {
-		printf("Tarantool/%s %s\n", mod_name, tarantool_version());
+		printf("Tarantool/Box %s\n", tarantool_version());
 		printf("Target: %s\n", BUILD_INFO);
 		printf("Build options: %s\n", BUILD_OPTIONS);
 		printf("Compiler: %s\n", COMPILER_INFO);
-		printf("CFLAGS:%s\n", COMPILER_CFLAGS);
+		printf("C_FLAGS:%s\n", TARANTOOL_C_FLAGS);
 		return 0;
 	}
 
@@ -574,7 +705,7 @@ main(int argc, char **argv)
 			panic("access(\"%s\"): %s", cat_filename, strerror(errno));
 			exit(EX_OSFILE);
 		}
-		return mod_cat(cat_filename);
+		return box_cat(cat_filename);
 	}
 
 	gopt_arg(opt, 'c', &cfg_filename);
@@ -606,7 +737,7 @@ main(int argc, char **argv)
 		strcat(cfg_filename_fullpath, cfg_filename);
 	}
 
-	cfg_out = tbuf_alloc(eter_pool);
+	cfg_out = tbuf_new(eter_pool);
 	assert(cfg_out);
 
 	if (gopt(opt, 'k')) {
@@ -649,8 +780,15 @@ main(int argc, char **argv)
 	if (cfg.username != NULL) {
 		if (getuid() == 0 || geteuid() == 0) {
 			struct passwd *pw;
+			errno = 0;
 			if ((pw = getpwnam(cfg.username)) == 0) {
-				say_syserror("getpwnam: %s", cfg.username);
+				if (errno) {
+					say_syserror("getpwnam: %s",
+						     cfg.username);
+				} else {
+					say_error("User not found: %s",
+						  cfg.username);
+				}
 				exit(EX_NOUSER);
 			}
 			if (setgid(pw->pw_gid) < 0 || setuid(pw->pw_uid) < 0 || seteuid(pw->pw_uid)) {
@@ -684,10 +822,9 @@ main(int argc, char **argv)
 	if (gopt(opt, 'I')) {
 		init_storage = true;
 		initialize_minimal();
-		mod_init();
-		next_lsn(recovery_state, 1);
-		confirm_lsn(recovery_state, 1);
-		snapshot_save(recovery_state, mod_snapshot);
+		box_init();
+		set_lsn(recovery_state, 1);
+		snapshot_save(recovery_state, box_snapshot);
 		exit(EXIT_SUCCESS);
 	}
 
@@ -719,34 +856,50 @@ main(int argc, char **argv)
 	atexit(tarantool_free);
 
 	ev_default_loop(EVFLAG_AUTO);
-
 	initialize(cfg.slab_alloc_arena, cfg.slab_alloc_minimal, cfg.slab_alloc_factor);
 	replication_prefork();
 
 	signal_init();
 
-	tarantool_L = tarantool_lua_init();
-	mod_init();
-	tarantool_lua_load_cfg(tarantool_L, &cfg);
-	admin_init();
-	replication_init();
 
-	/*
-	 * Load user init script.  The script should have access
-	 * to Tarantool Lua API (box.cfg, box.fiber, etc...) that
-	 * is why script must run only after the server was fully
-	 * initialized.
-	 */
-	tarantool_lua_load_init_script(tarantool_L);
-
-	prelease(fiber->gc_pool);
-	say_crit("log level %i", cfg.log_level);
-	say_crit("entering event loop");
-	if (cfg.io_collect_interval > 0)
-		ev_set_io_collect_interval(cfg.io_collect_interval);
-	ev_now_update();
-	start_time = ev_now();
-	ev_loop(0);
+	@try {
+		tarantool_L = tarantool_lua_init();
+		box_init();
+		memcached_init(cfg.bind_ipaddr, cfg.memcached_port);
+		tarantool_lua_load_cfg(tarantool_L, &cfg);
+		/*
+		 * init iproto before admin and after memcached:
+		 * recovery is finished on bind to the primary port,
+		 * and it has to happen before requests on the
+		 * administrative port start to arrive.
+		 * And when recovery is finalized, memcached
+		 * expire loop is started, so binding can happen
+		 * only after memcached is initialized.
+		 */
+		iproto_init(cfg.bind_ipaddr, cfg.primary_port,
+			    cfg.secondary_port);
+		admin_init(cfg.bind_ipaddr, cfg.admin_port);
+		replication_init(cfg.bind_ipaddr, cfg.replication_port);
+		session_init();
+		/*
+		 * Load user init script.  The script should have access
+		 * to Tarantool Lua API (box.cfg, box.fiber, etc...) that
+		 * is why script must run only after the server was fully
+		 * initialized.
+		 */
+		tarantool_lua_load_init_script(tarantool_L);
+		prelease(fiber->gc_pool);
+		say_crit("log level %i", cfg.log_level);
+		say_crit("entering event loop");
+		if (cfg.io_collect_interval > 0)
+			ev_set_io_collect_interval(cfg.io_collect_interval);
+		ev_now_update();
+		start_time = ev_now();
+		ev_loop(0);
+	} @catch (tnt_Exception *e) {
+		[e log];
+		panic("%s", "Fatal error, exiting loop");
+	}
 	say_crit("exiting loop");
 	/* freeing resources */
 	return 0;
