@@ -21,14 +21,12 @@ void test_ping()
 		0x0, 0x0, 0x0, 0x0,    0x0, 0x0, 0x0, 0x0,
 		0x0, 0x0, 0x0, 0x0,    0x1, 0x0, 0x0, 0x0,
 		0x4, 0x1, 0x0, 0x0, 0x0 };
-
 	tnt_io_send_raw(TNT_SNET_CAST(t), (char*)message, sizeof(message), 1);
 	t->wrcnt++;
-
 	struct tnt_iter i;
-	tnt_iter_stream(&i, t);
+	tnt_iter_reply(&i, t);
 	tnt_next(&i);
-	struct tnt_reply *r = TNT_ISTREAM_REPLY(&i);
+	struct tnt_reply *r = TNT_IREPLY_PTR(&i);
 	printf("return_code: %"PRIu32"\n", r->code); /* =0 */
 	tnt_iter_free(&i);
 }
@@ -37,23 +35,20 @@ void test_ping()
  * https://bugs.launchpad.net/tarantool/+bug/702397 "If SELECT
  * request specifies tuple count 0, no error"
  */
-
 void test_bug702397()
 {
 	const char message[]= {
 		0x11, 0x0, 0x0, 0x0,    0x14, 0x0, 0x0, 0x0,    0x0, 0x0, 0x0, 0x0,
 		0x0, 0x0, 0x0, 0x0,     0x0, 0x0, 0x0, 0x0,     0x0, 0x0, 0x0, 0x0,
 		0xff, 0xff, 0xff, 0xff, 0x0, 0x0, 0x0, 0x0 };
-
 	tnt_io_send_raw(TNT_SNET_CAST(t), (char*)message, sizeof(message), 1);
 	t->wrcnt++;
-
 	struct tnt_iter i;
-	tnt_iter_stream(&i, t);
+	tnt_iter_reply(&i, t);
 	tnt_next(&i);
-	struct tnt_reply *r = TNT_ISTREAM_REPLY(&i);
+	struct tnt_reply *r = TNT_IREPLY_PTR(&i);
 	printf("return_code: %s, %s\n",
-	       tnt_errcode_str(r->code >> 8), r->error);
+	       tnt_errcode_str(TNT_REPLY_ERR(r)), r->error);
 	tnt_iter_free(&i);
 }
 
@@ -61,7 +56,6 @@ void test_bug702397()
  * https://bugs.launchpad.net/tarantool/+bug/702399
  * ERR_CODE_ILLEGAL_PARAMS is returned when there is no such key
  */
-
 void test_bug702399()
 {
 	const char message[]= {
@@ -71,16 +65,36 @@ void test_bug702399()
 		0xff, 0xff, 0xff, 0xff,
 		0x1, 0x0, 0x0, 0x0,     0x1, 0x0, 0x0, 0x0,
 		0x4,    0x1, 0x0, 0x0, 0x0 };
-
 	tnt_io_send_raw(TNT_SNET_CAST(t), (char*)message, sizeof(message), 1);
 	t->wrcnt++;
-
 	struct tnt_iter i;
-	tnt_iter_stream(&i, t);
+	tnt_iter_reply(&i, t);
 	tnt_next(&i);
-	struct tnt_reply *r = TNT_ISTREAM_REPLY(&i);
+	struct tnt_reply *r = TNT_IREPLY_PTR(&i);
 	printf("return_code: %s, %s\n",
-	       tnt_errcode_str(r->code >> 8), r->error);
+	       tnt_errcode_str(TNT_REPLY_ERR(r)), r->error);
+	tnt_iter_free(&i);
+}
+
+/** A test case for Bug#1009992
+ * https://bugs.launchpad.net/tarantool/+bug/1009992
+ * ER_ILLEGAL_PARAMS is returned on bad operation id
+ */
+void test_bug1009992()
+{
+	struct tnt_header h = {
+		.type = 12345678, /* bad operation */
+		.len = 0,
+		.reqid = 0
+	};
+	tnt_io_send_raw(TNT_SNET_CAST(t), (char*)&h, sizeof(h), 1);
+	t->wrcnt++;
+	struct tnt_iter i;
+	tnt_iter_reply(&i, t);
+	tnt_next(&i);
+	struct tnt_reply *r = TNT_IREPLY_PTR(&i);
+	printf("return_code: %s, %s\n",
+	       tnt_errcode_str(TNT_REPLY_ERR(r)), r->error);
 	tnt_iter_free(&i);
 }
 
@@ -99,6 +113,7 @@ int main()
 	test_ping();
 	test_bug702397();
 	test_bug702399();
+	test_bug1009992();
 
 	tnt_stream_free(t);
 	return 0;
